local util = require"yoke.snap.util"
local concat,insert = table.concat,table.insert

local req = table.class('http.Request')
local parse = req:subclass('parse')
local hdrs = {}
local content = {}
local param = {}

function hdrs.parse(data)
  local headers = {}
  for key, val in string.gmatch(data, '([^%c%s:]+):%s+([^\n]+)') do
    key = string.lower(key)
    headers[key] = val
  end
  return headers
end

function hdrs.read(data, pos)
  local from, to = string.find(data,"\r\n\r\n", pos, true)
  if from then
    return hdrs.parse(string.sub(data, pos, from-1)), to+1
  else
    return nil, pos
  end
end

function hdrs.file(headers)
  local disp = headers["content-disposition"] or ""
  local attrs = {}
  for attr, val in string.gmatch(disp, ';%s*([^%s=]+)="(.-)"') do
    attrs[attr] = val
  end
  return attrs.name, attrs.filename and string.match(attrs.filename,"[/\\]?([^/\\]+)$")
end

function content.read(data, boundary, pos)
  local line = concat{"\r\n",boundary}
  local from, to = string.find(data, line, pos, true)
  if from then
    return string.sub(data, pos, from-1), from-pos, to+1
  else
    return nil, 0, pos
  end
end

function content.file(file,name,size, headers)
  local value = {contents = file, name = name, size = size}
  for key, val in pairs(headers) do
    if key ~= "content-disposition" then
      value[key] = val
    end
  end
  return value
end

local function iter(data, boundary)
  local state = {}
  _,state.pos = string.find(data, boundary, 1, true)
  state.pos = state.pos + 1
  return function (state,_)
    local headers, name, filename, value, size
    headers, state.pos = hdrs.read(data, state.pos)
    if headers then
      name, filename = hdrs.file(headers)
      if filename then
        value, size, state.pos = content.read(data, boundary,state.pos)
        value = content.file(value, filename, size, headers)
      else
        value, size, state.pos = content.read(data, boundary,state.pos)
      end
    end
    return name, value
  end, state
end

function parse:qs(qs)
  if type(qs) == "string" then
    local decode = util.url.decode
    for key, val in string.gmatch(qs, "([^&=]+)=([^&=]*)&?") do
      param.put(self.super.get, decode(key), decode(val))
    end
  elseif qs then
    error("request: error in query string")
  end
end

function parse:multipart(data)
  local boundary = string.match(self.env.contype or "", "boundary%=(.-)$")
  boundary = concat{"--",tostring(boundary)}
  for name, value in iter(data, boundary) do
    param.put(self.super.post, name, value)
  end
end

function parse:post()
  local input = self.input
  local contype = self.env.contype
  local conlen = tonumber(self.env.conlen) or 0
  if string.find(contype, "x-www-form-urlencoded", 1, true) then
    parse:qs(input:read(conlen) or "")
  elseif string.find(contype, "multipart/form-data", 1, true) then
    if conlen > 0 then
      parse:multipart(input:read(conlen) or "")
    end
  else
    self.body = input:read(len) or ""
  end
end

function parse:cookie()
  local cookies = string.gsub(";" .. (self.env.cookies or "") .. ";","%s*;%s*", ";")
  setmetatable(self.cookies, { __index = function (tab, name)
    local cookie = string.match(cookies,concat{";",name,"=(.-);"})
    cookie = util.url.decode(cookie)
    rawset(tab, name, cookie)
    return cookie
  end})
end

function param.put(tbl, name, value)
  if not tbl[name] then
    tbl[name] = value
  else
    local t = type(tbl[name])
    if t == "table" then
      insert(tbl[name], value)
    else
      tbl[name] = {tbl[name], value}
    end
  end
end

function param.join(get,post,out)
  setmetatable(out, { __index = function(pars, name)
      local val = get[name] or post[name]
      if val then
        rawset(pars, name, val)
      end
      return val
  end})
  for key,val in pairs(get) do _ = out[key] end --for rawset
end

function req:initialize(env)
  params = {}
  req.env          = {}
  req.env.method   = env.REQUEST_METHOD
  req.env.info     = env.PATH_INFO
  req.env.path     = env.APP_PATH
  req.env.qs       = env.QUERY_STRING
  req.env.script   = env.SCRIPT_NAME
  req.env.root     = env.DOCUMENT_ROOT
  req.env.contype  = env.CONTENT_TYPE
  req.env.conlen   = env.CONTENT_LENGTH
  req.env.cookies  = env.HTTP_COOKIE

  req.input        = env.input --stream
  req.cookies      = {}
  req.get          = {}
  req.post         = {}
  req.params       = {}

  parse:qs(req.env.qs)
  parse:post()
  parse:cookie()
  param.join(self.get,self.post,self.params)
end

return req

